from __future__ import annotations

import json
import pandas as pd
from functools import cached_property
from nandai.api import NandAIClient
from nandai.constants import DEFAULT_BATCH_SIZE
from nandai.constants import DEFAULT_DF_COLUMNS
from nandai.utils import replace_with_correction
import itables as it


class NandValidator:

    @cached_property
    def client(self):
        return NandAIClient()

    @cached_property
    def itables(self):
        it.options.css = """
        .nand-source { text-decoration: line-through; background-color: #ffeef0; color: #24292e; }
        .nand-correction { font-style: oblique; background-color: #e6ffed; color: #24292e; }
        """

        it.init_notebook_mode(all_interactive=False, warn_if_call_is_superfluous=False)
        return it

    async def validate(
            self,
            df: pd.DataFrame,
            columns: list[str] = DEFAULT_DF_COLUMNS,
            context: str | list[str] = [],
            batch_size: int = DEFAULT_BATCH_SIZE,
    ) -> pd.DataFrame:
        """Validate prompt-response pairs with context."""
        df.columns = df.columns.str.strip()
        res = await self.client.batch_validate(df, context=context, batch_size=batch_size, columns=columns)

        def replace(row: pd.Series):
            res: str = row['response']
            corr = json.loads(row['Correction'])
            row['CorrectedResponse'] = replace_with_correction(
                res,
                corr,
            )
            return row

        return res.apply(replace, axis=1)


    def display(self, df: pd.DataFrame, **kwargs):
        """Display validation result."""
        if df.empty:
            return

        for col in ['response', 'Correction']:
            if col not in df.columns:
                raise ValueError(f"`{col}` column not found in dataframe")

        def highlight(row: pd.Series):
            res: str = row['response']
            corr = json.loads(row['Correction'])
            row['response'] = replace_with_correction(
                res,
                corr,
                lambda x, y: f'<i class="nand-source">{x}</i><i class="nand-correction">{y}</i>',
            )
            return row

        df = df.apply(highlight, axis=1)
        kwargs = {'classes': 'display', **kwargs}
        self.itables.show(df, **kwargs)



import pytest
from pytest_mock import MockFixture
from unittest.mock import AsyncMock


@pytest.fixture
def mock_http_post(mocker: MockFixture) -> AsyncMock:
    json_mock = mocker.MagicMock()
    json_mock.json = mocker.MagicMock()
    mock = AsyncMock()
    mock.return_value = json_mock
    mocker.patch('nandai.api.http_client.post', mock)
    return json_mock.json



from __future__ import annotations

import json
import pandas as pd
import pytest

from nandai.api import NandAIClient


validate_response = json.loads("""
[
  {
    "factuality": true,
    "reasonings": [
      {
        "reasoning": "The given text is factual because the evidence provided states that 'a=1'.",
        "error": "None",
        "correction": "N/A",
        "reference": [
          "Additional Context"
        ],
        "factuality": true,
        "source": "the value of a is 1",
        "claim": "The value of a is 1"
      }
    ]
  }
]
""")


@pytest.mark.asyncio
async def test_client_validate(mock_http_post):
    mock_http_post.return_value = validate_response

    client = NandAIClient()
    res = await client.batch_validate(pd.DataFrame({
        'prompt': ['what value is a'],
        'response': ['the value of a is 1'],
    }), context='a=1')

    assert res['result'][0] == True
    assert res['raw_result'][0] == validate_response



from __future__ import annotations
import asyncio
import pytest
from nandai.utils import batch_async
from nandai.utils import replace_with_correction


async def async_func(data):
    await asyncio.sleep(0.01)
    return f"{data}"


@pytest.mark.asyncio
async def test_batch_async():
    data = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    batch_size = 3

    expected_results = [
        "1", "2", "3", "4", "5", "6", "7", "8", "9", "10"
    ]

    results = await batch_async(data, async_func, batch_size)

    assert results == expected_results


def test_replace_with_correction():
    text = 'Kjelsberg was born in Oslo, he was a Norwegian politician.'
    corr = {"Kjelsberg was born in Oslo": "Kjelsberg was born in Svelvik, Vestfold"}
    assert replace_with_correction(text, {}) == text
    assert replace_with_correction(text, corr) == 'Kjelsberg was born in Svelvik, Vestfold, he was a Norwegian politician.'
    assert replace_with_correction(text, corr, lambda x, y: 'www') == 'www, he was a Norwegian politician.'



from __future__ import annotations
import json
import httpx
import pandas as pd
import os

from nandai.constants import DEFAULT_BATCH_SIZE
from nandai.constants import DEFAULT_DF_COLUMNS
from nandai.utils import batch_async


transport = httpx.AsyncHTTPTransport(retries=2)
http_client = httpx.AsyncClient(transport=transport, timeout=120)


class NandAIClient:
    def __init__(
            self,
            url: str | None = None,
            api_key: str | None = None,
    ):
        self.url = url or os.getenv('NANDAI_API_URL') or 'https://llmeval.nand.ai/api'
        self.api_key = api_key or os.getenv('NANDAI_API_KEY') or ''

        if not self.url:
            raise ValueError('NANDAI_API_URL is not set as an environment variable or passed as an argument')

        if not self.api_key:
            raise ValueError('NANDAI_API_KEY is not set as an environment variable or passed as an argument')

    async def batch_validate(
            self,
            input: pd.DataFrame,
            columns: list[str] = DEFAULT_DF_COLUMNS,
            context: str | list[str] = [],
            batch_size: int = DEFAULT_BATCH_SIZE,
    ) -> pd.DataFrame:
        data = self._process_data(input, columns)
        context = self._process_additional_context(context)

        async def func(row: pd.Series):
            res = await self.validate(
                prompts=[row['prompt']],
                responses=[row['response']],
                context=context if 'context' not in row else row['context'] + '\n\n' + context,
               )
            row['NandScore'] = res[0]['NandScore']
            row['FactualInaccuraciesFound'] = res[0]['FactualInaccuraciesFound']
            row['Correction'] = res[0]['Correction']
            return row

        rows = [row for _, row in data.iterrows()]
        return pd.DataFrame(await batch_async(rows, func, batch_size)).sort_values(by='NandScore')

    def _process_data(self, data: pd.DataFrame, columns: list[str]) -> pd.DataFrame:
        """Prepare data for validation. Context column is optional."""
        ctx_col = 'context' if len(columns) == 2 else columns[2]
        columns = columns[:2]
        if ctx_col in data.columns:
            columns += [ctx_col]
        return data[columns].rename(columns=dict(zip(columns, ['prompt', 'response', 'context'])))

    def _process_additional_context(self, context: str | list[str]) -> str:
        if isinstance(context, str):
            context = [context]
        return '\n\n'.join(context)

    async def validate(self, prompts: list[str], responses: list[str], context: str):
        res = await http_client.post(
            url= f'{self.url}/llmeval',
            json={
                'prompt': prompts,
                'response': responses,
                'context': context,
                'excludes': [],
            },
            headers={
                'x-api-key': self.api_key,
            },
        )
        res = res.json()
        return [
            {
                'NandScore': r['nandscore'],
                'FactualInaccuraciesFound': not r['factuality'],
                'Correction': json.dumps(r['correction'])
            } for r in res
        ]



"""Use nandai to detect and fix LLM hallucinations.

The most commonly used functions are:
  - nandai.validate — validate one or multiple prompt-response pairs with optional context
"""

__version__ = "0.1.8"

from nandai.validator import NandValidator

_validator = NandValidator()
validate = _validator.validate
display = _validator.display

__all__ = (
    "__version__",
    "validate",
    "display",
)



import asyncio
import typer
import pandas as pd

try:
    from dotenv import load_dotenv
    load_dotenv('.env.local')
except:
    pass

from nandai.validator import NandValidator


cli = typer.Typer()
validator = NandValidator()


@cli.command()
def validate_one(prompt: str, response: str, context: str):
    """Validate a prompt-response pair with context."""
    print(f"Prompt: `{prompt}`, response: `{response}`, context: `{context}`")
    print("Running validation...")
    res = asyncio.run(validator.validate(
        pd.DataFrame({'prompt': [prompt], 'response': [response]}), context=context,
    ))
    print(f"\nResult: {res}")


@cli.command()
def validate(
        data_file: str,
        columns: str = typer.Option('prompt,response,context', '--columns', help="columns in data file, `context` is optional"),
        context: str = typer.Option('', '--context', '-c', help="additional context string that applies to all prompt-response pairs"),
        batch_size: int = 5,
        output: str = typer.Option('', '--output', '-o', help="Output file path"),
):
    """
    Validate prompt-response pairs from data file, supports 'csv' and 'jsonl' files.

    :return: File with additional columns: `NandScore`, `FactualInaccuraciesFound`, `Correction` (json string)
    """
    print(f"Data file: `{data_file}`, additional context: `{context}`")

    if data_file.endswith('.csv'):
        df = pd.read_csv(data_file)
    elif data_file.endswith('.jsonl') or data_file.endswith('.json'):
        df = pd.read_json(data_file, lines=True)
    else:
        raise ValueError(f"Unsupported data file type: `{data_file}`")

    cols = [c.strip() for c in columns.split(',')]
    print("Running batch validation...")
    res = asyncio.run(validator.validate(df, context=context, batch_size=batch_size, columns=cols))

    if output:
        print(f"Saving result to `{output}`")
        if output.endswith('.jsonl') or output.endswith('.json'):
            res.to_json(output, orient='records', lines=True)
        elif output.endswith('.csv'):
            res.to_csv(output)
        else:
            raise ValueError(f"Unsupported output file type: `{output}`")
    else:
        print(f"\nResult: {res}")


if __name__ == "__main__":
    cli()



from __future__ import annotations

import asyncio
import re
from tqdm import tqdm
from typing import Callable
from typing import Any


async def batch_async(data: list[Any], async_func: Callable, batch_size: int):
    res = []
    with tqdm(total=len(data)) as pbar:
        for i in range(0, len(data), batch_size):
            batch = data[i:i+batch_size]
            tasks = [async_func(d) for d in batch]
            res += await asyncio.gather(*tasks)
            pbar.update(len(batch))

    return res


def replace_with_correction(
        text: str,
        correction: dict[str, str],
        replace_func: Callable[[str, str], str] = lambda _, y: y,
):
    match_regex = re.compile('(' + "|".join(re.escape(m) for m in correction) + ')')
    parts = re.split(match_regex, text)
    return ''.join([
        replace_func(p, correction[p])
        if p in correction else p for p in parts
    ])
